unit SortableHashList;

(*                                                                             *
             THIS UNIT IS STILL UNDER CONSTRUCTION!                            *
             TODO: Separate ShortCompare_xx and ShortPos_xx to an own unit     *
                   Improve i386 algorithms similar to x86_64                   *
                                                                               *
                                                                               *
  Name:      SortableHashList                                                  *
                                                                               *
  Objective: * Fast HashList ShortString-Key sorting functionality with a      *
               descendant of Free Pascals ContNrs.TFPHashList.                 *
             * Fast case-sensitive and -insensitive ShortString Compare        *
               functions similar to SysUtils.CompareStr and CompareText        *
               in order to accelerate ShortString comparison and sorting.      *
                                                                               *
  Author:    Ruediger Walter <rue.walter@web.de>                               *
                                                                               *
  Version:   0.4.7 / 2015-03-05                                                *
                                                                               *
  Features:  TSortableHashList is a descendant of TFPHashList with the         *
             ability of sorting its keys and to hand them over to StringLists. *
             By nature of the hash algorithms a hashlist cannot be sorted      *
             internally and does not benefit from a sorted order in terms      *
             of searchtime. So the aim of sorting hashlist keys is basically   *
             the hand-over to a user interface, e.g. a ListBox. Typically      *
             this will be done via a Stringlist.                               *
             The sorting of TSortableHashList does not change anything inside  *
             the hashlist - the data themselves are not touched. Instead the   *
             basic idea of TSortableFPHashList is to provide an additional     *
             array to the hashlist, which contains pointers to the hashlists   *
             keys in sorted order.                                             *
                                                                               *
             TFPHashList is designed for speed, thus it's keys are limited     *
             to ShortStrings (if you have to deal with AnsiString keys         *
             instead, you should use one of the several TFPCustomHashTable-    *
             descendants).                                                     *
             The Free pascal RTL does not offer full specialized ShortString   *
             compare functions, it's "standard" functions SysUtils.CompareStr  *
             (case-sensitive) and CompareText (case-insensitive) are designed  *
             to deal with AnsiStrings. Any sorting procedure relies heavily    *
             on string comparison functions and will be significantly slowed   *
             down, if the compare function in use needs string type            *
             conversion on every single function call.                         *
             The CompareShortString(In)Sensitive functions are designed to     *
             avoid this string conversion and to deal efficiently with the     *
             internal structure of ShortStrings. Using them instead of the     *
             SysUtils standard routines speeds the hashList sorting up by      *
             a factor of about 3 to 6 (depending on case-sensitivity, data     *
             distribution and CPU).                                            *
                                                                               *
  Usage:     Constructor Create / Property CaseSensitivity: Sorting can be     *
             done either case-sensitive or case-insensitive. Parameters for    *
             creation of TSortableHashList and the property CaseSensitivity    *
             are either "CaseSensitive" or "CaseInsensitive".                  *
                                                                               *
             Procedure Sort / Property SortArray: The Sort procedure does not  *
             affect the internal data structure of the hashlist. First it      *
             writes the PShortString-Pointers of the hashlist keys into a      *
             dynamic PShortString-Array named "SortArray" in the indexed       *
             order of the hashlist. When Sort is done, those Shortstring-      *
             Pointers are redistributed in a sorted order.                     *
             Please note: There is no automatic update of SortArray, so if     *
             you change the hashlists content, SortArray will be out of date   *
             and you have to sort the list again. You should never write       *
             anything to SortArray, there's also no need to initiate or reset  *
             SortArray. Please keep in mind the specific rules of dealing      *
             with hashlist keys: E.g. an instruction like                      *
             >   SortArray[i]^ := 'I dont understand HashLists';               *
             will corrupt the hashlist data structure (in this case you have   *
             to use (TFPHashList.)Rename and call the Sort procedure again).   *
                                                                               *
             Procedure SortToStrings(ST:TStrings; LoadOption:TSHLLoadOption):  *
             Clears the stringlist ST, calls Sort and copies the hashlist      *
             keys in sorted order to the stringlist. There is no need to set   *
             ST.Sorted = true, this would only slow down the loading process.  *
             LoadOption can take the values "NoObjects" or "WithObjects".      *
             With "NoObjects" only the keys are copied to the stringlist,      *
             otherwise pointers to the hashlist items will be written to       *
             ST.Objects additionally. Please note that TStrings expects        *
             TObject-Pointers, while the hashlist accepts any kind of          *
             pointers. Adding those pointers to the stringlist is done via     *
             typecast and without any type check - you should know what you're *
             doing in the case of non-TObject-pointers (e.g. avoid setting     *
             ST.OwnObjects to true in this case).                              *
                                                                               *
             Function CompareShortString(In)Sensitive: Those functions are     *
             ShortString counterparts to the SysUtils-functions CompareStr     *
             and CompareText. They are not convenient for use with "normal"    *
             strings (which are mostly interpreted as AnsiStrings by FPC) -    *
             they explicitely use the internal representation of ShortStrings  *
             and will not work with AnsiStrings. Similar to their SysUtils     *
             counterparts they work on a byte-by-byte comparison basis and     *
             do not support WideChars, UTF-8 etc.                              *
             Compare Results are: S1 < S2 : Result < 0                         *
                                  S1 = S2 : Result = 0                         *
                                  S1 > S2 : Result > 0                         *
             S1 < S2 means:  (S1[n] < S2[n] on mismatch or                     *
                             Length(S1) < Length(S2) if all chars matched.     *
             Please note that the precise values besides equal, < or > are     *
             undefined.                                                        *
                                                                               *
  Limits:    * ShortStrings only                                               *
             * No automatic Sort update                                        *
                                                                               *
  Credits:   TSortableHashList.QuickSort was adapted from TFPGHashTable.Sort,  *
             and the Compare routines have been inspired by analyzing their    *
             RTL counterparts. So the real merits are on the side of the       *
             Free Pascal RTL authors!                                          *
                                                                               *
  Copyright (C) 2014 by Ruediger Walter <rue.walter@web.de>                    *
                                                                               *
  This library is licensed on the same Modifyed LGPL as Free Pascal RTL and    *
  Lazarus LCL. See the file COPYING.FPC, included in your FPC/Lazarus          *
  distribution, for details about the copyright.                               *
                                                                               *
  This program is distributed in the hope that it will be useful, but WITHOUT  *
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or        *
  FITNESS FOR A PARTICULAR PURPOSE.                                            *
                                                                               *)



{$mode objfpc} {$H-}

interface

uses
  Classes, SysUtils, Contnrs;


type
  TSHLSortArray   = array of PShortString;
  TSHLSensitivity = (CaseSensitive, CaseInsensitive);
  TSHLLoadOption  = (NoObjects, WithObjects);

  TShortCompareFunc = function(const S1, S2: ShortString): SmallInt;


  { TSortableHashList }

  TSortableHashList = class(TFPHashList)
  private
    FSortArray   : TSHLSortArray;
    FSensitivity : TSHLSensitivity;
    FCompFunc    : TShortCompareFunc;
    procedure   SetSensitivity(ASensitivity : TSHLSensitivity);
    procedure   DoQuickSort(L, R: integer);
  protected
    procedure   ResetSortArray;
    property    CompFunction : TShortCompareFunc read FCompFunc write FCompFunc;
  public
    constructor Create(ASensitivity : TSHLSensitivity);
    destructor  Destroy; override;
    procedure   Sort;
    procedure   SortToStrings(ST : TStrings; LoadOption : TSHLLoadOption = NoObjects);
    property    SortArray   : TSHLSortArray   read FSortArray;
    property    Sensitivity : TSHLSensitivity read FSensitivity write SetSensitivity;
  end;


function ShortCompareSensitive(const S1, S2: ShortString) : SmallInt;
function ShortCompareInsensitive(const S1, S2: ShortString) : SmallInt;


implementation


{ ***************************** TSortableHashList ***************************** }

constructor TSortableHashList.Create(ASensitivity : TSHLSensitivity);
begin
  inherited Create;
  SetSensitivity(ASensitivity);
end;

destructor TSortableHashList.Destroy;
begin
  SetLength(FSortArray, 0);
  inherited Destroy;
end;

procedure TSortableHashList.SetSensitivity(ASensitivity : TSHLSensitivity);
begin
  { The usual "if Sensitivity = FCaseOption then exit;" will trigger a SIGSEGV error
    by skipping the FCompFunc-assignemnt from Create, if Sensitivity = CaseSensitive
    (which internally is handled as zero, to which FCaseOption is initiated on creation)!
    Otherwise the following lines must be doubled in TSortableHashList.Create. }
  FSensitivity := ASensitivity;
  if ASensitivity = CaseSensitive then
    FCompFunc := @ShortCompareSensitive
  else FCompFunc := @ShortCompareInsensitive;
end;

procedure TSortableHashList.DoQuickSort(L, R : integer);
var
  I, J, P        : Integer;
  tmp, PivotItem : PShortString;
begin
  repeat
    I := L;
    J := R;
    P := (L + R) div 2;
    repeat
      PivotItem := FSortArray[P];
      while (P <> I) and (FCompFunc(PivotItem^, FSortArray[I]^) > 0) do Inc(I);
      while (P <> J) and (FCompFunc(PivotItem^, FSortArray[J]^) < 0) do Dec(J);
      if I <= J then
      begin
        tmp := FSortArray[i];
        FSortArray[I] := FSortArray[J];
        FSortArray[J] := tmp;
        if P = I then P := J
        else if P = J then P := I;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then DoQuickSort(L, J);
    L := I;
  until I >= R;
end;

procedure TSortableHashList.ResetSortArray;
var
  n : integer;
begin
  SetLength(FSortArray, Count);
  for n := 0 to Count - 1 do
    FSortArray[n] := PShortString(@Strs[List^[n].StrIndex]);
end;

procedure TSortableHashList.Sort;
begin
  ResetSortArray;
  if (Count < 2) then exit;
  DoQuickSort(0, Count-1);
end;

procedure TSortableHashList.SortToStrings(ST : TStrings; LoadOption : TSHLLoadOption);
var
  i : integer;
begin
  if not Assigned(ST) then exit;
  ST.Clear;
  Sort;
  if LoadOption = WithObjects then
    for i := 0 to Count - 1 do
      ST.AddObject(FSortArray[i]^, TObject(Find(FSortArray[i]^)))
  else for i := 0 to Count - 1 do
    ST.Add(FSortArray[i]^);
end;




{ ************************* ShortCompare(In)Sensitive ************************* }


{ ************************* X86_64 Assembler versions ************************* }

{ $CODEALIGN PROC=16}
{ $CODEALIGN LOOP=16}

{$OPTIMIZATION REGVAR}
{$ASMMODE att}
{$ifdef CPUX86_64}

{ Function ShortCompareSensitive():
  For case-sensitive compare full register block comparison can be used. But because
  we need a lower/greater-decision (if not equal; the precise result values besides
  that are undefined) the string blocks must be reversed in the registers into a Big-Endian
  order (the first string mismatch from left to right must be at the most significant
  register position). For the same reason the specialized CPU string instructions
  (REPE/REPZ CMPSQ/D/W) cannot be used (and using them bytewise (CMPSB) is way too slow).

Register allocation:
_____________________________________________________________________________________________________
|   RAX                 |   RCX                 |   RDX                 |     RSI     |     RDI     |
|    :   EAX            |    :   ECX            |    :   EDX            |             |             |
|    :    :     AX      |    :    :     CX      |    :    :     DX      |             |             |
|    :    :  AH  :  AL  |    :    :  CH  :  CL  |    :    :  DH  :  DL  |             |             |
|----:----:------:------|----:----:------:------|----:----:------:------|-------------|-------------|
|     S1^ (Q/L/W/B)     | -- : -- : LengthCount |     S2^ (Q/L/W/B)     |    @S1[n]   |    @S2[n]   |
----------------------------------------------------------------------------------------------------}

function ShortCompareSensitive(const S1, S2 : ShortString) : SmallInt;
Begin
  asm
  { Preliminaries: Register preparation and Length evaluation / Length result assignment }
             movq   S1, %rsi         // @S1 -> RSI
             movq   S2, %rdi         // @S2 -> RDI
             movzbw (%rsi), %ax      // Length(S1) -> AX
             movzbw (%rdi), %dx      // Length(S2) -> DX
             movw   %ax, %cx         // Length(S1) in AX -> CX
             subw   %dx, %ax         // Length result (Length(S1) in AX - Length(S2) in DX) -> AX
             cmovaw %dx, %cx         // if Length(S1) in CX > Length(S2) in DX then Length(S2) -> CX
             movw   %ax, result      // Assign Length result
             addq   $1, %rsi         // Increment @S1 in RSI to the first Char
             addq   $1, %rdi         // Increment @S2 in RDI to the first Char
  { Block loading and comparison in the RAX/RDX registers }
    .LLoad8: cmpw   $8, %cx          // Length >= 8 chars ?
             jb     .LLoad4          // if length < 8 then try 4 byte loading
             movq   (%rsi), %rax     // 8 Chars(S1) -> RAX
             movq   (%rdi), %rdx     // 8 Chars(S2) -> RDX
             subw   $8, %cx          // Update Counter
             addq   $8, %rsi         // Update Pointer S1
             addq   $8, %rdi         // Update Pointer S2
             cmpq   %rdx, %rax       // Compare all 8 Chars S1 in RAX - S2 in RDX
             je     .LLoad8          // if equal, discard registers and goto next 8 byte loading
             bswapq %rax             // RAX: Little Endian -> Big Endian
             bswapq %rdx             // RDX: Little Endian -> Big Endian
             cmpq   %rdx, %rax       // compare for mismatch result assignment
             ja     .LMisRa          // if Mismatch and S1 > S2 then goto LMisRa where Result := +1
             jb     .LMisRb          // if Mismatch and S1 < S2 then goto LMisRb where Result := -1
    .LLoad4: cmpw   $4, %cx          // Length >= 4 Chars ?
             jb     .LLoad2          // if Length < 4 then try 2 byte loading
             movl   (%rsi), %eax     // 4 Chars(S1) -> EAX
             movl   (%rdi), %edx     // 4 Chars(S2) -> EDX
             subw   $4, %cx          // Update Counter
             addq   $4, %rsi         // Update Pointer S1
             addq   $4, %rdi         // Update Pointer S2
             cmpl   %edx, %eax       // Compare all 4 Chars S1 in EAX - S2 in EDX
             je     .LLoad2          // if equal, discard registers and goto next 2 byte loading
             bswapl %eax             // EAX: Little Endian -> Big Endian
             bswapl %edx             // EDX: Little Endian -> Big Endian
             cmpl   %edx, %eax       // compare for mismatch result assignment
             ja     .LMisRa          // if Mismatch and S1 > S2 then goto LMisRa where Result := +1
             jb     .LMisRb          // if Mismatch and S1 < S2 then goto LMisRb where Result := -1
    .LLoad2: cmpw   $2, %cx          // Length >= 2 Chars ?
             jb     .LLoad1          // if Length < 8 then try 1 byte loading
             movw   (%rsi), %ax      // 2 Chars(S1) -> AX
             movw   (%rdi), %dx      // 2 Chars(S2) -> DX
             subw   $2, %cx          // Update Counter
             addq   $2, %rsi         // Update Pointer S1
             addq   $2, %rdi         // Update Pointer S2
             cmpw   %dx, %ax         // Compare all 2 Chars S1 in AX - S2 in DX
             je     .LLoad1          // if equal, discard registers and goto last 1 byte loading
             xchgb  %ah, %al         // AX: Little Endian -> Big Endian
             xchgb  %dh, %dl         // DX: Little Endian -> Big Endian
             cmpw   %dx, %ax         // compare for mismatch result assignment
             ja     .LMisRa          // if Mismatch and S1 > S2 then goto LMisRa where Result := +1
             jb     .LMisRb          // if Mismatch and S1 < S2 then goto LMisRb where Result := -1
    .LLoad1: cmpw   $1, %cx          // Length = 1 Char ? Reached only, if this is the last char
             jb     .LExit           // if Length = 0 then it's done, Length result is already assigned
             movb   (%rsi), %al      // Last Char(S1) -> AL
             movb   (%rdi), %dl      // Last Char(S2) -> DL, no update of CX and pointers needed any more
             cmpb   %dl, %al         // compare for mismatch result assignment
             je     .LExit           // If equal, it's done, exit, Length result is already assigned
             jb     .LMisRb          // if Mismatch and S1 < S2 then goto LMisRb where Result := -1
  { Mismatch result assignment. As the mismatch result value besides <0 or >0 is not defined, calculation
    could be replaced by a simple assignment of +1 or -1. In this case this is more efficient, because
    otherwise the exact byte position of the mismatch must be determined and shifted to AL/DL first. }
    .LMisRa: movw   $1, result       // if S1 > S2 then Result := 1
             jmp    .LExit           // Done...
    .LMisRb: movw   $-1, result      // if S1 < S2 then Result := -1
    .LExit:
  end ['rax', 'cx', 'rdx', 'rsi', 'rdi'];
end;


{ Function ShortCompareInsensitive():
  Other than CompareShortStringSensitive the insensitive compare function doesn't
  use Endian swapping. This is because of the necessity to inspect and possibly
  upcase every single Byte in case of a block mismatch. This can't be done blockwise,
  we have to work from left to right in the string, which means from right to left
  in the registers, instead. Doing so it's the more convenient way to have the registers
  in Little Endian order.

Register allocation:
_____________________________________________________________________________________________________
|   RAX                 |   RCX                 |   RDX                 |     RSI     |     RDI     |
|    :   EAX            |    :   ECX            |    :   EDX            |             |             |
|    :    :     AX      |    :    :     CX      |    :    :     DX      |             |             |
|    :    :  AH  :  AL  |    :    :  CH  :  CL  |    :    :  DH  :  DL  |             |             |
|----:----:------:------|----:----:------:------|----:----:------:------|-------------|-------------|
|     S1^ (Q/L/W/B)     | -- : -- : LengthCount |     S2^ (Q/L/W/B)     |    @S1[n]   |    @S2[n]   |
----------------------------------------------------------------------------------------------------}

function ShortCompareInsensitive(const S1, S2 : ShortString) : SmallInt;
Begin
  asm
  { Preliminaries: Register preparation and Length evaluation / Length result assignment }
             movq   S1, %rsi         // @S1 -> RSI
             movq   S2, %rdi         // @S2 -> RDI
             movzbw (%rsi), %ax      // Length(S1) -> AX
             movzbw (%rdi), %dx      // Length(S2) -> DX
             movw   %ax, %cx         // Length(S1) in AX -> CX
             subw   %dx, %ax         // Length result (Length(S1) in AX - Length(S2) in DX) -> AX
             cmovaw %dx, %cx         // if Length(S1) in CX > Length(S2) in DX then Length(S2) -> CX
             movw   %ax, result      // Assign Length result
             addq   $1, %rsi         // Increment @S1 in RSI to the first Char
             addq   $1, %rdi         // Increment @S2 in RDI to the first Char
             xorq   %rax, %rax       // Clear RAX (needed for block comparison, if length < 8)
             xorq   %rdx, %rdx       // Clear RDX
  { Block loading into the RAX/RDX registers }
    .LLoad8: cmpw   $8, %cx          // Length >= 8 chars ?
             jb     .LLoad4          // if length < 8 then try 4 byte loading
             movq   (%rsi), %rax     // 8 Chars(S1) -> RAX
             movq   (%rdi), %rdx     // 8 Chars(S2) -> RDX
             subw   $8, %cx          // Update Counter
             addq   $8, %rsi         // Update Pointer S1
             addq   $8, %rdi         // Update Pointer S2
             cmpq   %rdx, %rax       // Compare all 8 Chars S1 in RAX - S2 in RDX
             je     .LLoad8          // if equal, discard registers and goto next 8 byte loading
             jmp    .LComp4          // On mismatch goto 4 Byte Comparison
    .LLoad4: cmpw   $4, %cx          // Length >= 4 Chars ? Only reached, if remaining length < 8
             jb     .LLoad2          // if Length < 8 then try 2 byte loading
             movl   (%rsi), %eax     // 8 Chars(S1) -> RAX
             movl   (%rdi), %edx     // 8 Chars(S2) -> RDX
             subw   $4, %cx          // Update Counter
             addq   $4, %rsi         // Update Pointer S1
             addq   $4, %rdi         // Update Pointer S2
             jmp    .LComp4          // Goto 4 Byte Comparison
    .LLoad2: cmpw   $2, %cx          // Length >= 2 Chars ? Only reached, if remaining length < 4
             jb     .LLoad1          // if Length < 8 then try 1 byte loading
             movw   (%rsi), %ax      // 2 Chars(S1) -> AX
             movw   (%rdi), %dx      // 2 Chars(S2) -> DX
             subw   $2, %cx          // Update Counter
             addq   $2, %rsi         // Update Pointer S1
             addq   $2, %rdi         // Update Pointer S2
             jmp    .LComp2          // Goto 2 Byte Comparison
    .LLoad1: cmpw   $1, %cx          // Length 0 or 1 Char ? Only reached, if remaining length < 2
             jb     .LExit           // if Length = 0 then it's done, Length result is assigned
             movb   (%rsi), %al      // Last Char(S1) -> AL
             movb   (%rdi), %dl      // Last Char(S2) -> DL
             subw   $1, %cx          // Update Counter, updating Pointers S1/S2 is obsolete now
             jmp    .LComp1          // Goto 1 Byte Comparison
  { Multi char comparison inside the RAX/RDX registers }
    .LComp8: cmpq   %rdx, %rax       // Compare all 8 Chars in RAX/RDX
             je     .LLoad8          // if equal, discard registers and goto next 8 byte loading
    .LComp4: cmpl   %edx, %eax       // Compare lower 4 Chars, there must be a mismatch in RAX/RDX
             jne    .LComp2          // if not equal, mismatch may be in EAX/EDX, try lower 2 byte comparison
             shrq   $32, %rax        // else if equal, shift right RAX/RDX by 4 bytes
             shrq   $32, %rdx        // Now a mismatch may be in EAX/EDX
    .LComp2: cmpw   %dx, %ax         // Compare lower 2 Chars
             jne    .LComp1          // if not equal, mismatch is in AX/DX, try lowest byte comparison
             shrq   $16, %rax        // else if equal, shift right RAX/RDX by 2 bytes
             shrq   $16, %rdx        // Now a mismatch may be in AX/DX
    .LComp1: cmpb   %dl, %al         // Compare lowest Char in AL/DL
             jne    .LUpAL           // if not equal, try lowest byte upcase comparison
             shrq   $8, %rax         // else if equal, shift right RAX/RDX by 1 byte
             shrq   $8, %rdx         // Now a mismatch may be in AL/DL, so go on to upcase comparison
             cmpq   %rdx, %rax       // shortens, if there is no mismatch any more
             je     .LLoad8          // jump to LLoad8!
  { Single Char upcase comparison on case-sensitive mismatch }
    .LUpAL:  cmpb   $97, %al         // if Char in AL < "a" then ...
             jb     .LUpDL           //   ... goto upcasing Char in DL
             cmpb   $122, %al        // if Char in AL > "z" then ...
             ja     .LMisM           //   ... goto upcasing Char in DL
             xorb   $32, %al         // Now Char in AL is in ['a'..'z'], upcase by dec(AL, 32)
    .LUpDL:  cmpb   $97, %dl         // if Char in DL < "a" then ...
             jb     .LUpCmp          //   ... goto upcased char comparison
             cmpb   $122, %dl        // if Char in DL > "z" then ...
             ja     .LMisM           //   ... goto upcased char comparison
             xorb   $32, %dl         // Now Char in DL is in ['a'..'z'], upcase by dec(DL, 32)
    .LUpCmp: cmpb   %dl, %al         // Comparing the (possibly) upcased chars again
             je     .LComp8          // if equal, go back to block comparison (shr here is less efficient)
  { Mismatch result assigment. When reached, the mismatch is definitely located in AL/DL }
    .LMisM:  xorb   %ah, %ah         // if not equal, discard the upper bytes ...
             xorb   %dh, %dh         //   ... and prepare AX/DX registers for SmallInt result assignment
             subw   %dx, %ax         // Mismatch result -> AX
             movw   %ax, result      // Assign mismatch function result
    .LExit:
  end ['rax', 'cx', 'rdx', 'rsi', 'rdi'];
end;

// TODO: rework algorithm according to x86_64 for better performance
{$else (ifndef CPUX86_64)}



{ ************************** i386 Assembler versions ************************** }


{$ifdef CPUI386}

{Function ShortCompareSensitive():

Register allocation:
_____________________________________________________________________________________________________
|      EAX              |      ECX              |      EDX              |     ESI     |     EDI     |
|               AX      |       :       CX      |       :       DX      |             |             |
|       :   AH  :   AL  |       :   CH  :   CL  |       :   DH  :   DL  |             |             |
|----:----:------:------|----:----:------:------|----:----:------:------|-------------|-------------|
|      S1^ (L/W/B)      | -- : -- : LengthCount |      S2^ (L/W/B)      |    @S1[n]   |    @S2[n]   |
----------------------------------------------------------------------------------------------------}

function ShortCompareSensitive(const S1, S2 : ShortString): SmallInt;
begin
  asm
{ Preliminaries: Register preparation }
           movl   S1, %edi         // @S1 -> EDI
           movl   S2, %esi         // @S2 -> ESI
           xor    %ecx, %ecx       // Clear ECX, used as counter register
{ Length evaluation }
           movzbw (%edi), %ax      // Length-Byte(S1) -> AX
           movzbw (%esi), %dx      // Length-Byte(S2) -> DX
           inc    %edi             // Increment @S1 in EDI to the first Char
           inc    %esi             // Increment @S2 in ESI to the first Char
           movw   %dx, %cx         // Length(S2) -> CX
           cmp    %dx, %ax         // Compare (Length(S1) in AX - Length(S2) in DX)
           jae    .L1              // if Length(S2) <= Length(S1) then CX is okay
           movw   %ax, %cx         //   else Length(S1) -> CX;
    .L1:   sub    %dx, %ax         // (Length(S1) - Length(S2)) -> AX
           movw   %ax, Result      // (Length(S1) - Length(S2)) in AX -> Result (needed on match)
{ Begin with 4 Byte substring comparison }
    .L2:   cmp    $4, %cx          // Length >= 4 ?
           jb     .L3              // If Length < 4 then goto L4 (2 Byte comparison)
           movl   (%edi), %eax     // 4 Byte String S1 from EDI -> EAX (faster than LODSD/L)
           movl   (%esi), %edx     // 4 Byte String S2 from ESI -> EDX
           bswap  %eax             // EAX: Little Endian -> Big Endian
           bswap  %edx             // EDX: Little Endian -> Big Endian
           cmp    %edx, %eax       // Compare 4 Byte EAX - EDX = S1 - S2
           ja     .L6              // if Mismatch and S1 > S2 then goto L6 where Result := +1
           jb     .L5              // if Mismatch and S1 < S2 then goto L5 where Result := -1
           add    $4, %edi         // inc(String Pointer EDI, 4)
           add    $4, %esi         //   and inc(String Pointer ESI, 4)
           sub    $4, %cx          // dec(Length in CX, 4)
           jmp    .L2              // try again 8 Byte Compare
{ 2 Byte substring comparison, there can't be more than 3 Chars left }
    .L3:   cmp    $2, %cx          // Length >= 2 ?
           jb     .L4              // If Length < 2 then goto L4 (last Byte comparison)
           movw   (%edi), %ax      // 2 Byte String S1 from EDI -> AX
           movw   (%esi), %dx      // 2 Byte String S2 from ESI -> DX
           cmp    %dl, %al         // Compare AL - DL Byte first due to Little Endian
           ja     .L6              // if Mismatch and S1 > S2 then goto L6 where Result := +1
           jb     .L5              // if Mismatch and S1 < S2 then goto L5 where Result := -1
           cmp    %dh, %ah         // Compare AH - DH Byte next (= S1 - S2)
           ja     .L6              // if Mismatch and S1 > S2 then goto L6 where Result := +1
           jb     .L5              // if Mismatch and S1 < S2 then goto L5 where Result := -1
           sub    $2, %cx          // dec(Length in CX, 4)
           jz     .Lx              // Exit if no Byte left (all matched, result is already assigned)
           add    $2, %edi         // inc(String Pointer EDI, 2)
           add    $2, %esi         //   and inc(String Pointer ESI, 2)
{ Last Byte comparison, there must be 0 or 1 Byte left if reached }
    .L4:   jecxz  .Lx              // No Byte left, all matched, Length result is already assigned
           movb   (%edi), %al      // Last Byte of S1 from EDI -> AL
           movb   (%esi), %dl      // Last Byte of S2 from ESI -> DL
           cmp    %dl, %al         // Compare last Byte, no counter update needed any more
           je     .Lx              // Last Byte okay, all matched, Length result is already assigned
           ja     .L6              // if Mismatch and S1 > S2 then goto L6 where Result := +1
{ Mismatch Result assignment }
    .L5:   movw   $-1, result      // if S1 < S2 then Result := -1
           jmp    .Lx              // Done...
    .L6:   movw   $1, result       // if S1 > S2 then Result := 1
    .Lx:                           // Exit
  end ['eax', 'ecx', 'edx', 'esi', 'edi'];
end;


{Function ShortCompareInsensitive():

Register allocation:
_____________________________________________________________________________________________________
|      EAX              |      ECX              |      EDX              |     ESI     |     EDI     |
|               AX      |       :       CX      |       :       DX      |             |             |
|       :   AH  :   AL  |       :   CH  :   CL  |       :   DH  :   DL  |             |             |
|----:----:------:------|----:----:------:------|----:----:------:------|-------------|-------------|
|      S1^ (L/W/B)      | -- : -- : LengthCount |      S2^ (L/W/B)      |    @S1[n]   |    @S2[n]   |
----------------------------------------------------------------------------------------------------}

function ShortCompareInsensitive(const S1, S2 : ShortString) : SmallInt;
Begin
  asm
{ Preliminaries: Register preparation }
           movl   S1, %edi         // @S1 -> EDI
           movl   S2, %esi         // @S2 -> ESI
           xor    %ecx, %ecx       // Clear ECX, used as counter register
           { Length evaluation }
           movzbw (%edi), %ax      // Length-Byte(S1) -> AX
           movzbw (%esi), %dx      // Length-Byte(S2) -> DX
           inc    %esi             // Increment @S1 in ESI to the first Char
           inc    %edi             // Increment @S2 in EDI to the first Char
           movw   %dx, %cx         // Length(S2) -> CX
           cmp    %dx, %ax         // Compare (Length(S1) in AX - Length(S2) in DX)
           jge    .L1              // if Length(S2) <= Length(S1) then CX is okay
           movw   %ax, %cx         //   else Length(S1) -> CX;
    .L1:   sub    %dx, %ax         // (Length(S1) - Length(S2)) -> AX
           movw   %ax, Result      // (Length(S1) - Length(S2)) in AX -> Result (needed on match)
{ Now begin with 4 Byte substring comparison }
    .L2:   cmp    $4, %cx          // Length >= 4 ?
           jb     .L3              // If Length < 4 then goto L4 (2 Byte comparison)
           movl   (%edi), %eax     // 4 Byte String S1 from EDI -> EAX (faster than LODSD/L)
           movl   (%esi), %edx     // 4 Byte String S2 from ESI -> EDX
           cmp    %edx, %eax       // Now compare whole 4 Byte substring (EAX - EDX) = (S1 - S2)
           jne    .L3a             // if Mismatch goto L3 without updating counters: Compare lower Word
           add    $4, %edi         // DWord has matched, now update counters
           add    $4, %esi         // inc String Pointers ESI & EDI by 4
           sub    $4, %cx          // dec Length in CX by 4
           jmp    .L2              // If matched then try again 4 Byte Compare
{ 2 Byte substring comparison, but there might be more than 3 Chars left! }
    .L3:   cmp    $2, %cx          // Length >= 2 ?
           jb     .L4              // If Length < 2 then goto L4 (last Byte comparison)
           movw   (%edi), %ax      // 2 Byte String S1 from EDI -> AX and inc(RSI, 4)
           movw   (%esi), %dx      // 2 Byte String S2 from ESI -> DX
           // Next 2 Bytes surely not out of String and loaded in AX/DX, but counters not yet updated
    .L3a:  cmp    %dx, %ax         // Compare 2 Byte AX - DX = S1 - S2
           jne    .L3b             // on Mismatch goto L4b without updating counters: Which Byte ?
           add    $2, %edi         // Word has matched, now update counters
           add    $2, %esi         // inc String Pointers ESI & EDI by 2
           subw   $2, %cx          // dec Length in CX by 2: Are there any Bytes left?
           jz     .Lx              // If not, it's done, Length comparison result is already assigned
           // The reason why it reached this point is either a mismatch step-down or because it's
           // near the end of the string. In the first case the mismatch must be in the next 2 Bytes
           jmp    .L3              // therefore if there are Bytes left, try again 2 Byte Compare
           // Here a mismatch is detected in AX/DX, but where is it?
    .L3b:  cmp    %dl, %al         // First compare the lower Byte
           jne    .Lc              // If mismatch in AL/DL goto .Lc, registers & flags are already set
           inc    %edi             // Now the mismatch must be in AH/DH,
           inc    %esi             //   so update counters,
           dec    %cx
           shr    $8, %ax          //   shift H -> L
           shr    $8, %dx
           jmp    .Lc              //   and goto .Lc for case match comparison
{ Single Byte comparison, here can't be anything else than no or the (unloaded) last Char left! }
    .L4:   jecxz  .Lx              // No Byte left, all matched, Length result is already assigned
           movb   (%edi), %al      // Next Byte from String S1 in EDI -> AL and inc(RSI, 4)
           movb   (%esi), %dl      // Next Byte from String S2 in ESI -> DL
           cmpb   %dl, %al         // Compare last Byte
           je     .Lx              // if matched, it's done, Length comparison result is already set
{ Case comparison: Here's a mismatch in AL/DL with correct counter positions set, but contrary
  to the position at Label .L4 we don't know whether it's the last Byte or not! So if there is a
  case match, we have to jump back to .L2 (DWord comparison). In fact, this is the only way
  a different-case match tracks back to block comparison. }
    .Lc:   cmpb   $97, %al         // Eliminating case differences by upcasing chars in AL/DL:
           jb     .Lc1             // if Char in AL < "a" ...
           cmpb   $122, %al        // ... or ...
           ja     .Lc1             // if Char in AL > "z" do nothing
           subb   $32, %al         // else dec(AL, 32)
    .Lc1:  cmpb   $97, %dl
           jb     .Lc2             // Now do the same with DL
           cmpb   $122, %dl
           ja     .Lc2
           subb   $32, %dl
    .Lc2:  cmp    %dl, %al         // Comparing the (possibly) upcased chars:
           jne    .L5              // On mismatch goto result assignment
           inc    %edi             // On match update counters
           inc    %esi             //   (In most cases this is not the end of the line!)
           dec    %cx
           jmp    .L2              // and jump back to .L2 DWord comparison
{ Mismatch Result assignment }
    .L5:   ja     .L6              // Mismatch: if S1 > S2 then goto L7, where Result := +1
           mov    $-1, result      // else Result := -1
           jmp    .Lx              // Done...
    .L6:   mov    $1, result       // Result := +1
    .Lx:                           // Exit
  end ['eax', 'ecx', 'edx', 'esi', 'edi'];
end;


{$else (ifndef (CPUX86_64 or CPUI386))}

{ ****************************** Pascal versions ****************************** }


function ShortCompareSensitive(const S1, S2 : ShortString) : SmallInt;
var
  c1, c2  : Byte;
  L1, L2,
  Cnt     : SizeInt;
  P1, P2  : PChar;
begin
  L1 := Length(S1);
  L2 := Length(S2);
  P1 := @S1[1];
  P2 := @S2[1];
  if L1 > L2 then Cnt := L2 else Cnt := L1;
  while cnt > 0 do
  begin
    c1 := byte(p1^);
    c2 := byte(p2^);
    if c1 <> c2 then
    begin
      Result := c1 - c2;
      Exit;
    end;
    Inc(P1); Inc(P2); Dec(Cnt);
  end;
  Result := L1 - L2;
end;

function ShortCompareInsensitive(const S1, S2 : ShortString) : SmallInt;
begin
  result := ShortCompareText(S1, S2);
end;

{$endif CPUI386}
{$endif CPUX86_64}

end.


